// $Id$
/**
 * Copyright (C) 2014 EDIT
 * European Distributed Institute of Taxonomy
 * http://www.e-taxonomy.eu
 *
 * The contents of this file are subject to the Mozilla Public License Version 1.1
 * See LICENSE.TXT at the top of this package for the full license terms.
 */
package eu.etaxonomy.taxeditor.webapp;

import java.io.File;
import java.io.IOException;
import java.net.ServerSocket;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;

import javax.management.InstanceNotFoundException;
import javax.management.MBeanException;
import javax.management.MBeanServerConnection;
import javax.management.MalformedObjectNameException;
import javax.management.ObjectName;
import javax.management.ReflectionException;
import javax.management.remote.JMXConnector;
import javax.management.remote.JMXConnectorFactory;
import javax.management.remote.JMXServiceURL;

import org.apache.log4j.Logger;
import org.eclipse.core.runtime.FileLocator;
import org.eclipse.core.runtime.Platform;
import org.eclipse.jetty.server.Server;
import org.eclipse.jetty.util.StringUtil;
import org.eclipse.jetty.util.preventers.AppContextLeakPreventer;
import org.eclipse.jetty.webapp.WebAppContext;
import org.osgi.framework.Bundle;

/**
 *
 * (Singleton) Server instance which manages a compatible cdmlib-webapp-war.
 * This is done by launching a jetty instance (using jetty-runner) as an
 * executed process. The choice of the external process over a more
 * preferable 'embedded jetty' instance is due to problems arising from the
 * class loading of classes (e.g. from hibernate core) which are needed
 * for both the webapp as well as the remoting client.
 *
 * @author cmathew
 * @date 23 Sep 2014
 *
 */

public class CDMServer {

    public static final Logger logger = Logger.getLogger(CDMServer.class);


    private final String dataSourceName;
    private final String host = "127.0.0.1";
    private int httpPort = 9090;
    private final String contextPath = "/";

    private File warFile;
    private Server server;


    public CDMServer(String dataSourceName, File dataSourcesFile) throws CDMEmbeddedServerException {
        if(StringUtil.isBlank(dataSourceName)) {
            throw new CDMEmbeddedServerException("DataSource name is not valid");
        }

        if(dataSourcesFile == null || !dataSourcesFile.exists()) {
            throw new CDMEmbeddedServerException("DataSource config file does not exist");
        }
        this.dataSourceName = dataSourceName;
        Bundle bundle = Platform.getBundle("eu.etaxonomy.taxeditor.webapp");
        URL warURL = bundle.getEntry("lib/cdmlib-remote-webapp.war");

        try {
            URL resolvedWarURL = FileLocator.toFileURL(warURL);
            // We need to use the 3-arg constructor of URI in order to properly escape file system chars
            URI resolvedURI = new URI(resolvedWarURL.getProtocol(), resolvedWarURL.getPath(), null);
            warFile = new File(resolvedURI);
            System.out.println("war url : " + warFile.getAbsolutePath());

            if(warFile == null || !warFile.exists()) {
                throw new CDMEmbeddedServerException("Cdmlib War file does not exist");
            }
        } catch (URISyntaxException use) {
            throw new CDMEmbeddedServerException(use);
        } catch (IOException ioe) {
            throw new CDMEmbeddedServerException(ioe);
        }

        System.setProperty("spring.profiles.active", "remoting");
        System.setProperty("cdm.beanDefinitionFile", dataSourcesFile.getAbsolutePath());
        System.setProperty("cdm.datasource", dataSourceName);

        httpPort = findFreePort();

        logger.warn("Starting server on port : " + httpPort);

        server = new Server(httpPort);

        server.addBean(new AppContextLeakPreventer());

        WebAppContext webapp = new WebAppContext();
        webapp.setContextPath(contextPath);
        webapp.setWar(warFile.getAbsolutePath());

        server.setHandler(webapp);
    }

    public String getDataSourceName() {
        return dataSourceName;
    }
    public String getHost() {
        return host;
    }

    public int getPort() {
        return httpPort;
    }

    public String getContextPath() {
        return contextPath;
    }

    private static int findFreePort() throws CDMEmbeddedServerException {
        ServerSocket socket = null;
        try {
            socket = new ServerSocket(0);
            socket.setReuseAddress(true);
            int port = socket.getLocalPort();
            try {
                socket.close();
            } catch (IOException e) {

            }
            return port;
        } catch (IOException e) {
        } finally {
            if (socket != null) {
                try {
                    socket.close();
                } catch (IOException e) {
                }
            }
        }
        throw new CDMEmbeddedServerException("Could not find a free TCP/IP port to start embedded Jetty HTTP Server on");
    }

    public void start(ICDMServerError cdmServerError) throws CDMEmbeddedServerException {
        start(true, cdmServerError);
    }

    public void start(boolean wait, final ICDMServerError cdmServerError) throws CDMEmbeddedServerException {

        if(server == null) {
            throw new CDMEmbeddedServerException("Server is already disposed");
        }

        if(server.isStarting()) {
            throw new CDMEmbeddedServerException("Server is starting");
        }

        if(server.isStarted()) {
            throw new CDMEmbeddedServerException("Server has started");
        }

        if(server.isRunning()) {
            throw new CDMEmbeddedServerException("Server is running");
        }

        if(server.isStopping()) {
            throw new CDMEmbeddedServerException("Server is currently stopping. Please re-try in about 10 seconds");
        }

        Thread serverThread = new Thread() {

            @Override
            public void run() {
                try {
                    server.start();
                    server.join();
                } catch (Throwable t) {
                    cdmServerError.handleError(t);
                }
            }
        };

        serverThread.start();

        if(wait) {
            while(!server.isStarted()) {}

        }
    }

    public boolean isAlive()  {
        return server.isRunning() || server.isStarting() || server.isStarted();
    }

    public boolean isStarted() {
        return server.isStarted();
    }

    public void stop() throws Exception {
        server.stop();
        server.destroy();
        server = null;
    }




    public static void stopServerViaJMX(int jmxPort) throws CDMEmbeddedServerException  {
        String JMX_URL = "service:jmx:rmi:///jndi/rmi://localhost:" + jmxPort + "/jmxrmi";
        logger.warn("Shutting down Jetty instance ... ");

        try {
            JMXConnector connector = JMXConnectorFactory.connect(new JMXServiceURL(JMX_URL), null);
            connector.connect(null);
            MBeanServerConnection connection = connector.getMBeanServerConnection();
            ObjectName objectName = new ObjectName("org.eclipse.jetty.server:type=server,id=0");
            connection.invoke(objectName, "stop", null, null);
            logger.warn("Shutdown command sent");
        } catch (InstanceNotFoundException e) {
            throw new CDMEmbeddedServerException(e);
        } catch (MBeanException e) {
            throw new CDMEmbeddedServerException(e);
        } catch (ReflectionException e) {
            throw new CDMEmbeddedServerException(e);
        } catch (IOException e) {
            throw new CDMEmbeddedServerException(e);
        } catch (MalformedObjectNameException e) {
            throw new CDMEmbeddedServerException(e);
        }
    }


}
